"""Libraries for data science"""

import typing
from typing import Any, Callable, Iterable, Optional

if typing.TYPE_CHECKING:  # for sphinx
    import numpy as np

    np.zeros(0)

from pyloggrid.Libs.singlethread_numpy import np

np.zeros(0)  # Do not remove - used to force singlethread_numpy to be imported before the other libraries when refactoring
import scipy.optimize


def filter_range(data: np.ndarray, start: float = 0, end: float = 1) -> np.ndarray:
    """Trims a 1D array from its first and last portions.

    This is a shortcut to ``data[int(size * start) : int(size * end)]``

    Args:
        data: the data to filter
        start: where to start in the values (ex: 0.2 -> we ommit the 20% first values)
        end: where to end in the values (ex: 0.8 -> we ommit the 20% last values)

    Returns:
        the new array *(if filtered, the shape is altered)*
    """
    N = data.size
    return data[int(N * start) : int(N * end)]


def mean(data: np.ndarray, ts: Optional[np.ndarray] = None, log: bool = False, start: float = 0, end: float = 1) -> tuple[np.ndarray, np.ndarray | None]:
    """Mean of ``data``. If ``ts`` is provided, the mean is weighted by the time intervals. If ``log``, the log mean is returned.

    Args:
        data: data to average
        ts: times associated with the data. Must be increasing.
        log: if True, only the mean avg is returned ``exp(mean(log(abs(data_weighted)))``, and the variance returns ``None``
        start: where to start in the values (ex: 0.2 -> we ommit the 20% first values)
        end: where to end in the values (ex: 0.2 -> we ommit the 20% last values)

    Returns:
        the mean and std value
    """

    weights = np.ones_like(data) if ts is None else np.gradient(ts)
    weights = filter_range(weights, start=start, end=end)
    data = filter_range(data, start=start, end=end)

    if log:
        return np.exp(np.average(np.log(np.abs(data)), weights=weights)), None

    avg = np.average(data, weights=weights)
    variance = np.sqrt(np.average((data - avg) ** 2, weights=weights))  # https://stackoverflow.com/questions/2413522/weighted-standard-deviation-in-numpy
    return avg, variance


def clamp_zero(data: np.ndarray, fill: Any = None) -> np.ndarray:
    """Replace all zero values with ``fill`` (defaults to ``min(data)/10``)

    Args:
        data: array to clamp
        fill: if specified, all zero values are filled with this

    Returns:
        The clamped array
    """
    data = np.array(data)
    if data.size == 0:
        return data
    if np.max(np.abs(data)) == 0:
        fill = np.nextafter(0, 1)

    data[data == 0] = max(np.min(np.abs(data[data != 0])) / 10, np.nextafter(0, 1)) if fill is None else fill

    return data


def fit(f: Callable, x: np.ndarray, y: np.ndarray, mask: np.ndarray | tuple[float, float] = None) -> tuple[float, ...]:
    """Fit ``(x,y)`` by the function ``f`` with an optional mask

    Args:
        f: fit function
        x
        y
        mask: an array mask, or a tuple (start, end)
    """
    if isinstance(mask, tuple):
        x = filter_range(x, mask[0], mask[1])
        y = filter_range(y, mask[0], mask[1])
    elif mask is not None:
        x = x[mask]
        y = y[mask]
    mask2 = np.isfinite(y)
    x = x[mask2]
    y = y[mask2]
    assert len(x) > 1, "Your fitted data must be of size >1, check your mask and NaN data"

    # noinspection PyUnresolvedReferences
    params, _ = scipy.optimize.curve_fit(f, x, y)
    return params


def powerlaw_fit(x: np.ndarray, y: np.ndarray, mask: np.ndarray | float = None, a: float = None) -> tuple[float, float]:
    """Power law fit ``Y = b * X ^ a``

    Args:
        x: X array
        y: Y array
        mask: mask to specify the relevant values of X, Y to use (same as :func:`fit`)
        a: optional fixed value for ``a``. If specified, only ``b`` is fitted

    Returns:
        the fit parameters ``b, a``
    """
    if a is None:
        lnb, a = fit(lambda logx, lnb, a: lnb + a * logx, np.log(x), np.log(y), mask)
    else:
        lnb = fit(lambda logx, lnb: lnb + a * logx, np.log(x), np.log(y), mask)
    return np.exp(lnb), a


def randcomplex_like(data: np.ndarray) -> np.ndarray:
    """Random array of complex numbers, see :func:`randcomplex`

    Args:
        data: array to copy the shape from

    Returns:
        random array
    """
    return randcomplex(data.shape)


def randcomplex(shape: Iterable) -> np.ndarray:
    """Random array of complex numbers, normally distributed in modulus and uniformly in angle

    Args:
        shape: return array shape

    Returns:
        random array
    """
    # noinspection PyTypeChecker
    return np.random.normal(size=shape) * np.exp(1.0j * np.random.uniform(0, 2 * np.pi, shape))


def rand_seeded_by_array(source: np.ndarray, seed: int) -> np.ndarray:
    r"""Returns a pseudorandom array of same shape as ``source`` where items with the same ``source`` will have the same value for a given seed.

    This is useful when creating pseudorandom arrays based on ``kx,ky,kz``, of different sizes, e.g. when changing the grid size but keeping the same pseudorandom forcing

    Warning:
        This is not optimised for speed, and should probably not be called every step
    """
    res = np.zeros_like(source)
    values, indexes = np.unique(source, return_index=True)

    for v in values:
        np.random.seed(int(seed + v) % (2**30))
        res[source == v] = np.random.random()

    return res


def randcomplex_seeded_by_array(source: np.ndarray, seed: int) -> np.ndarray:
    """Like :func:`rand_seeded_by_array`, but returns a complex array (uniform in a square)"""
    return rand_seeded_by_array(source, seed) + 1j * rand_seeded_by_array(source, seed + 1)


def ragged_array_to_array(data: np.ndarray | list, fill=np.nan) -> np.ndarray:
    """Transform an 1D array of 1D array of different sizes into an array of array of same sizes, i.e a classic 2D array.

    Smaller arrays are padded with ``fill``.

    If a classic 2D array is given, it does nothing.

    Args:
        data: array of array of different sizes
        fill: fill value

    Returns:
        padded 2D-array
    """
    if len(data) == 0:
        return np.array(data)
    len_max = len(max(data, key=len))
    for ite in range(len(data)):
        fill_array = np.empty(len_max - len(data[ite]))
        fill_array[:] = fill
        data[ite] = np.append(data[ite], fill_array)
    len_x = len(data)
    len_y = len(data[0])
    data = np.concatenate(data)
    data = data.reshape((len_x, len_y))
    return data


def logmean(data: np.ndarray, axis=None) -> np.ndarray:
    """Logarithmic mean along an axis, ignoring zero and nan values."""
    return np.exp(np.nanmean(np.log(clamp_zero(data, np.NaN)), axis=axis))



"""Solver for the log grids."""

from __future__ import annotations

import typing

if typing.TYPE_CHECKING:  # for Sphinx
    import numpy as np

    np.zeros(0)

from pyloggrid.Libs.singlethread_numpy import np

np.zeros(0)  # Do not remove - used to force singlethread_numpy to be imported before the other libraries when refactoring

import contextlib
import datetime
import logging
import os
import shutil
import sys
import time
from typing import Callable, Optional, Union

from rkstiff import etd35

from pyloggrid.Libs.custom_logger import setup_custom_logger
from pyloggrid.Libs.IOLib import load_step, read_json_settings, save_step, update_json_settings
from pyloggrid.Libs.misc import TimeTracker
from pyloggrid.Libs.singlethread_numpy import np
from pyloggrid.LogGrid.Grid import Grid

logger = setup_custom_logger(__name__, level=logging.INFO)


## A few useful functions


def fields_to_1D(fields: dict[str, np.ndarray], field_names: list[str]) -> np.ndarray:
    """
    Convert N-dimensional fields to 1D arrays

    Args:
        field_names: [name1, ...] name of the fields (order is important, as it it used to convert back from 1D)
        fields: N-dimensional named fields, grid-shaped

    Returns:
        1D ndarray containing all the appended 1D fields, as ordered by ``field_names``
    """
    each_field_linear = np.array([fields[arr_name].reshape(fields[arr_name].size) for arr_name in field_names])
    return each_field_linear.reshape(each_field_linear.size)


def D1_to_fields(linear: np.ndarray, field_names: list[str], field_shape: (int, ...)) -> dict[str, np.ndarray]:
    """Reverse of :func:`fields_to_1D`

    Converts one 1D array to a dict of the encoded fields in their original shape

    Args:
        linear: 1D ndarray of fields
        field_names: [name1, ...] name of the fields (order is important)
        field_shape: shape of the grid / fields

    Returns:
        the fields with their full shape
    """
    linears_with_fields = linear.reshape((len(field_names),) + field_shape)
    return {field_names[i]: linears_with_fields[i].reshape(field_shape) for i in range(len(field_names))}


class _CustIntegrator:
    """Abstract integrator class"""

    def __init__(
        self,
        equation_nl: Callable[[float, np.ndarray], np.ndarray],
        equation_l: Callable[[float, np.ndarray], np.ndarray],
        save_step: Callable[[float, float, np.ndarray], bool],
        init_t: float,
        solver_params: dict,
        y0: np.ndarray,
    ):
        self.equation_nl = equation_nl
        self.equation_l = equation_l
        self.save_step = save_step
        self.init_t = init_t

        self.y = np.array(y0)
        self.t = self.init_t
        self.status = "not started"
        self.stopsim = False  # to stop the simulation

        self.minh = solver_params.get("minh", 1e-16)

    def select_initial_step(self, t0: float, y0: np.ndarray, order: int, rtol: float = 1e-6, atol: float = 1e-6) -> float:
        """Empirically select a good initial step.
        The algorithm is described in [#]_.

        Args:
            t0: Initial value of the independent variable.
            y0: Initial value of the dependent variable.
            order: Error estimator order. It means that the error controlled by the algorithm is proportional to ``step_size ** (order + 1)``.
            rtol: Desired relative tolerance.
            atol: Desired absolute tolerance.

        Returns:
            ``h_abs`` = Absolute value of the suggested initial step.

        References:
            .. [#] E. Hairer, S. P. Norsett G. Wanner, "Solving Ordinary Differential Equations I: Nonstiff Problems", Sec. II.4.
        """
        norm = lambda x: np.linalg.norm(x) / x.size**0.5
        direction = 1
        if y0.size == 0:
            return np.inf

        scale = atol + np.abs(y0) * rtol
        scale_valid = np.abs(scale) != 0
        h0 = 1e-6
        f0 = self.equation_nl(t0, y0)
        dd0 = np.zeros(scale.shape, dtype="complex")
        dd0[scale_valid] = y0[scale_valid] / scale[scale_valid]
        d0 = norm(dd0)
        dd1 = np.zeros(dd0.shape, dtype="complex")
        dd1[scale_valid] = f0[scale_valid] / scale[scale_valid]
        d1 = norm(dd1)
        if d0 >= 1e-5 and d1 >= 1e-5:
            h0 = max(0.01 * d0 / d1 if np.isfinite(d1) else self.minh, self.minh)
        y1 = y0 + h0 * direction * f0
        f1 = self.equation_nl(t0 + h0 * direction, y1)
        dd2 = np.zeros(scale.shape, dtype="complex")
        dd2[scale_valid] = (f1 - f0)[scale_valid] / scale[scale_valid]
        d2 = norm(dd2) / h0

        if d1 <= 1e-15 and d2 <= 1e-15:
            h1 = max(1e-6, h0 * 1e-3)
        else:
            h1 = max((0.01 / max(d1, d2)) ** (1 / (order + 1)) if np.isfinite(d2) else self.minh, self.minh)

        return min(100 * h0, h1)

    # noinspection PyMissingOrEmptyDocstring
    def solve(self):
        raise NotImplementedError


class ViscDopri(_CustIntegrator):
    """DOPRI5-based solver made specifically for log grids.

    A nonlinear update step is decoupled from the linear update step.
    Special care is given to handling numerical (float-induced) random numerical errors which yield huge gradients.
    Based on scipy's DOPRI5 implementation.

    Args:
        equation_nl: linear-less update step
        equation_l: linear update step
        init_t: initial time
        y0: initial 1D array
        solver_params: = ``{"atol"[absolute tolerance]: 1e-6, "rtol"[relative tolerance]: 1e-4, "exact_viscous_splitting"[if True, use exact viscous splitting]: False}``
        dt_params: ``{dt0 = initial time step, dtmin, dtmax}``
    """

    def __init__(
        self,
        equation_nl: Callable[[float, np.ndarray], np.ndarray],
        equation_l: Callable[[float, np.ndarray], np.ndarray],
        save_step: Callable[[float, float, np.ndarray], bool],
        init_t: float,
        solver_params: dict,
        y0: np.ndarray,
        dt_params: dict[str, Optional[float]],
    ):
        super().__init__(equation_nl, equation_l, save_step, init_t, solver_params, y0)
        self.atol = solver_params.get("atol", 1e-6)
        self.rtol = solver_params.get("atol", 1e-4)
        self.exact_viscous_splitting = solver_params.get("exact_viscous_splitting", False)

        # Internal RK & DOPRI parameters
        self.K = np.empty((7, self.y.size), dtype=self.y.dtype)
        self.max_factor = 10
        self.min_factor = 0.2
        self.min_step = 1e-200
        self.error_exponent = -1 / 5
        self.safety = 0.9
        self.A = np.array(
            [
                [0, 0, 0, 0, 0],
                [1 / 5, 0, 0, 0, 0],
                [3 / 40, 9 / 40, 0, 0, 0],
                [44 / 45, -56 / 15, 32 / 9, 0, 0],
                [19372 / 6561, -25360 / 2187, 64448 / 6561, -212 / 729, 0],
                [9017 / 3168, -355 / 33, 46732 / 5247, 49 / 176, -5103 / 18656],
            ]
        )
        self.B = np.array([35 / 384, 0, 500 / 1113, 125 / 192, -2187 / 6784, 11 / 84])
        self.C = np.array([0, 1 / 5, 3 / 10, 4 / 5, 8 / 9, 1])
        self.E = np.array([-71 / 57600, 0, 71 / 16695, -71 / 1920, 17253 / 339200, -22 / 525, 1 / 40])

        # Initial step
        self.dt = dt_params["dt0"] if "dt0" in dt_params else self.select_initial_step(self.t, self.y, 4, self.rtol, self.atol)
        self.dtmin, self.dtmax = dt_params.get("dtmin"), dt_params.get("dtmax")
        logger.info(f"Initial step: dt={self.dt}")

    def rk_step(
        self, fun: Callable[[float, np.ndarray], np.ndarray], t: float, y: np.ndarray, h: float, A: np.ndarray, B: np.ndarray, C: np.ndarray, K: np.ndarray
    ) -> tuple[np.ndarray, np.ndarray]:
        """Perform a single Runge-Kutta step.
        This function computes a prediction of an explicit Runge-Kutta method and
        also estimates the error of a less accurate method.
        Notation for Butcher tableau is as in [#]_.

        Args:
            fun: Right-hand side of the system.
            t: Current time.
            y: Current state.
            h: Step to use.
            A: Coefficients for combining previous RK stages to compute the next stage. For explicit methods the coefficients at and above the main diagonal are zeros.
            B: Coefficients for combining RK stages for computing the final prediction.
            C: Coefficients for incrementing time for consecutive RK stages. The value for the first stage is always zero.
            K: Storage array for putting RK stages here. Stages are stored in rows. The last row is a linear combination of the previous rows with coefficients

        Returns: tuple ``(y_new, f_new)`` where ``y_new`` is the solution at ``t + h`` computed with a higher accuracy, and ``f_new`` is the derivative ``fun(t + h, y_new)``.

        References:
            .. [#] E. Hairer, S. P. Norsett G. Wanner, "Solving Ordinary Differential Equations I: Nonstiff Problems", Sec. II.4.
        """

        if self.exact_viscous_splitting:
            fun_old = fun

            # noinspection PyMissingOrEmptyDocstring
            def fun_new(t_, y_):
                dt = t_ - t
                visc = self.equation_l(t_, y_)
                return fun_old(t_, y_) * np.exp(-visc * dt)

            fun = fun_new

        K[0] = fun(t, y)
        for s, (a, c) in enumerate(zip(A[1:], C[1:]), start=1):
            dy = np.dot(K[:s].T, a[:s]) * h
            K[s] = fun(t + c * h, y + dy)

        y_new = y + h * np.dot(K[:-1].T, B)
        f_new = fun(t + h, y_new)

        K[-1] = f_new

        return y_new, f_new

    def rkdp(self, force_step: bool = False) -> tuple[np.ndarray, float, float]:
        """Performs one RKDP step"""

        # quick access
        atol, rtol = self.atol, self.rtol
        t = self.t
        y = self.y
        A, B, C, E = self.A, self.B, self.C, self.E

        h = max(self.dt, self.min_step)

        # try to step once
        step_accepted, step_rejected = False, False
        y_new, derivative_new, newh = None, None, None
        while not step_accepted:
            assert h >= self.min_step, f"Could not converge (step too small {h})"

            # force step if dt is too small compared to float precision. Not a great practice, but can "free" the stepper from tough landscapes.
            forced_step = False
            if self.dtmax is not None and h > self.dtmax:
                h = self.dtmax
            elif self.dtmin is not None and h < self.dtmin:
                h = self.dtmin
                forced_step = True
            if force_step:
                if h < t * 1e-15:
                    h = max(h, 3e-16 * t)
                    logger.warning(f"Forcing step, h/t={h / t}")
                    forced_step = True
            else:
                assert t + h > t, f"Can not add step {h} to time t={t}"

            # don't store derivative from prev step, because the linear step changes it in an unknown way
            y_new, derivative_new = self.rk_step(self.equation_nl, t, y, h, A, B, C, self.K)

            # Compute error
            scale = atol * max(np.max(np.abs(y)), np.max(np.abs(y_new))) + np.maximum(np.abs(y), np.abs(y_new)) * rtol
            # remove too small values
            absurd_vals = scale == 0
            scale[absurd_vals] = 1
            RK_diff = np.abs(np.dot(self.K.T, E))
            RK_diff[absurd_vals] = 0
            error_norm = np.max(RK_diff * h / np.abs(scale))  # was np.norm
            # Check error value, and adjust time steps
            if error_norm < 1 or forced_step:
                step_accepted = True
                if error_norm == 0:
                    factor = self.max_factor
                else:
                    # noinspection PyTypeChecker
                    factor = max(self.min_factor, min(self.max_factor, self.safety * error_norm**self.error_exponent))

                if step_rejected:
                    factor = min(1.0, factor)

                newh = h * factor
            else:
                step_rejected = True
                h *= max(self.min_factor, self.safety * error_norm**self.error_exponent)

            if forced_step:
                newh = max(t / 1e15 / 3, newh)
        # Perform linear update
        visc = self.equation_l(t, y_new)
        y_new = y * np.exp(visc * h) + (y_new - y) * np.sum([np.exp(visc * h * (1 - c)) * b for b, c in zip(self.B, self.C)], axis=0)
        # y_new = y_new * np.exp(visc * h)  # old way of updating: more stable but less accurate
        self.stopsim = self.save_step(t + h, h, y_new)

        return y_new, h, newh

    def solve(self) -> None:
        """Main solving loop.

        The solver runs until an interruption is fired by the update."""
        self.status = "running"
        while not self.stopsim:
            newy, useddt, newdt = self.rkdp()
            self.y = newy
            self.dt = newdt
            self.t += useddt

        self.status = "finished"


class ETD4RK(_CustIntegrator):
    r"""Based on Cox & Matthews (2002). Exponential time differencing for stiff systems. *Journal of Computational Physics*, 176(2), 430-455.

    Warning:
        Only designed for constant-in-time viscosities

    .. note:: See ``Utils/thresholds.py`` in the source code for the determination of the thresholds used in computing Taylor series

    Args:
        equation_nl: linear-less update step
        equation_l: linear update step
        init_t: initial time
        y0: initial 1D array
        dt_params: ``{dt0 = initial time step, dtmin, dtmax}``
    """

    def __init__(
        self,
        equation_nl: Callable[[float, np.ndarray], np.ndarray],
        equation_l: Callable[[float, np.ndarray], np.ndarray],
        save_step: Callable[[float, float, np.ndarray], bool],
        init_t: float,
        y0: np.ndarray,
        dt_params: dict[str, Optional[float]],
    ):
        super().__init__(equation_nl, equation_l, save_step, init_t, {}, y0)

        # Initial step
        self.dt = dt_params["dt0"]
        logger.info(f"Initial step: dt={self.dt}")

    @staticmethod
    def dopri_step(
        fun: Callable[[float, np.ndarray], np.ndarray], t: float, y: np.ndarray, h: float, A: np.ndarray, B: np.ndarray, C: np.ndarray, K: np.ndarray
    ) -> tuple[np.ndarray, np.ndarray]:
        """Perform a single Runge-Kutta step.
        This function computes a prediction of an explicit Runge-Kutta method and
        also estimates the error of a less accurate method.
        Notation for Butcher tableau is as in [#]_.

        Args:
            fun: Right-hand side of the system.
            t: Current time.
            y: Current state.
            h: Step to use.
            A: Coefficients for combining previous RK stages to compute the next stage. For explicit methods the coefficients at and above the main diagonal are zeros.
            B: Coefficients for combining RK stages for computing the final prediction.
            C: Coefficients for incrementing time for consecutive RK stages. The value for the first stage is always zero.
            K: Storage array for putting RK stages here. Stages are stored in rows. The last row is a linear combination of the previous rows with coefficients

        Returns: tuple ``(y_new, f_new)`` where ``y_new`` is the solution at ``t + h`` computed with a higher accuracy, and ``f_new`` is the derivative ``fun(t + h, y_new)``.
        References:
            .. [#] E. Hairer, S. P. Norsett G. Wanner, "Solving Ordinary Differential Equations I: Nonstiff Problems", Sec. II.4.
        """
        K[0] = fun(t, y)
        for s, (a, c) in enumerate(zip(A[1:], C[1:]), start=1):
            dy = np.dot(K[:s].T, a[:s]) * h
            K[s] = fun(t + c * h, y + dy)

        y_new = y + h * np.dot(K[:-1].T, B)
        f_new = fun(t + h, y_new)

        K[-1] = f_new

        return y_new, f_new

    def rk(self) -> np.ndarray:
        """Performs one Rk step"""
        # quick access
        t = self.t
        y = self.y  # u_n
        h = self.dt
        visc = self.equation_l(t, y)  # c
        hc = h * visc
        e = np.exp(hc / 2)
        e2 = np.exp(hc)
        F = self.equation_nl

        # perform RK. We need to take care of hc<<1 terms, where the precision of the exp becomes too bad to cancel Taylor terms
        F0 = F(t, y)
        a = np.zeros_like(y)
        mask0 = np.abs(hc) > 1e-6
        a[mask0] = y[mask0] * e[mask0] + (e[mask0] - 1) * F0[mask0] / visc[mask0]  # (26)
        a[~mask0] = y[~mask0] * e[~mask0] + F0[~mask0] * h / 2
        Fa = F(t + h / 2, a)
        b = np.zeros_like(y)
        b[mask0] = y[mask0] * e[mask0] + (e[mask0] - 1) * Fa[mask0] / visc[mask0]  # (27)
        b[~mask0] = y[~mask0] * e[~mask0] + Fa[~mask0] * h / 2
        Fb = F(t + h / 2, b)
        c = np.zeros_like(y)
        c[mask0] = a[mask0] * e[mask0] + (e[mask0] - 1) * (2 * Fb[mask0] - F0[mask0]) / visc[mask0]  # (28)
        c[~mask0] = a[~mask0] * e[~mask0] + (2 * Fb[~mask0] - F0[~mask0]) * h / 2
        Fc = F(t + h, c)

        el0, el1, el2, y_new = np.zeros_like(y), np.zeros_like(y), np.zeros_like(y), np.zeros_like(y)
        mask1 = np.abs(hc) > 1e-3
        el0[mask1] = F0[mask1] * (-4 - hc[mask1] + e2[mask1] * (4 - 3 * hc[mask1] + hc[mask1] ** 2)) / (h**2 * visc[mask1] ** 3)
        el0[~mask1] = F0[~mask1] * h / 6
        el1[mask1] = 2 * (Fa[mask1] + Fb[mask1]) * (2 + hc[mask1] + e2[mask1] * (-2 + hc[mask1])) / (h**2 * visc[mask1] ** 3)
        el1[~mask1] = 2 * (Fa[~mask1] + Fb[~mask1]) * h / 6
        mask2 = np.abs(hc) > 6e-3
        el2[mask2] = Fc[mask2] * (-4 - 3 * hc[mask2] - hc[mask2] ** 2 + e2[mask2] * (4 - hc[mask2])) / (h**2 * visc[mask2] ** 3)
        el2[~mask2] = Fc[~mask2] * h / 6
        y_new = y * e2 + el0 + el1 + el2  # (29)

        self.stopsim = self.save_step(t + h, h, y_new)

        return y_new

    def solve(self) -> None:
        """Main solving loop. The solver runs until an interruption is fired by the update."""
        self.status = "running"
        while not self.stopsim:
            newy = self.rk()
            self.y = newy
            self.t += self.dt

        self.status = "finished"


class ETD35(_CustIntegrator):
    """Based on https://github.com/whalenpt/rkstiff

    Warning:
        Only supports diagonal (and constant-in-time) linear terms for now

    Args:
        equation_nl: linear-less update step
        equation_l: linear update step
        init_t: initial time
        solver_params: ``{"rtol"[relative tolerance]: 1e-2, "adapt_cutoff"[Limits values used in the computation of the suggested step size to those with |u| > adapt_cutoff*max(|u|)]: 1e-2, "minh"[minimum time step]: 1e-16}``
        y0: initial 1D array
        dt_params: ``{dt0 = initial time step, dtmin, dtmax}``
    """

    def __init__(
        self,
        equation_nl: Callable[[float, np.ndarray], np.ndarray],
        equation_l: Callable[[float, np.ndarray], np.ndarray],
        save_step: Callable[[float, float, np.ndarray], bool],
        init_t: float,
        solver_params: dict,
        y0: np.ndarray,
        dt_params: dict[str, Optional[float]],
    ):
        super().__init__(equation_nl, equation_l, save_step, init_t, solver_params, y0)
        self.rtol = solver_params.get("rtol", 1e-2)
        self.adapt_cutoff = solver_params.get("adapt_cutoff", 1e-2)

        # Initial step
        self.dt = dt_params["dt0"] if "dt0" in dt_params else self.select_initial_step(self.t, self.y, 4)
        self.dt_suggested = self.dt
        logger.info(f"Initial step: dt={self.dt}")

    def rk(self) -> tuple[np.ndarray, float, float]:
        """Performs one Rk step

        Returns:
            tuple ``(y_new, h, h_new_suggested)``
        """
        # quick access
        t = self.t
        y = self.y

        solver = etd35.ETD35(
            linop=self.equation_l(t, y), NLfunc=lambda y_: self.equation_nl(t, y_), epsilon=self.rtol, adapt_cutoff=self.adapt_cutoff, minh=self.minh
        )
        y_new, h, h_new_suggested = solver.step(y, self.dt_suggested)

        self.stopsim = self.save_step(t + h, h, y_new)
        return y_new, h, h_new_suggested

    def solve(self) -> None:
        """Main solving loop. The solver runs until an interruption is fired by the update."""
        self.status = "running"
        while not self.stopsim:
            self.y, self.dt, self.dt_suggested = self.rk()
            self.t += self.dt

        self.status = "finished"


class SolverInterruptedError(InterruptedError):
    """custom error thrown to change grid size"""

    ...


class Solver:
    """Generic top-level object to handle solving log-grid equations

    Args:
        D: dimension of physical space (usually 1, 2 or 3)
        l_params: parameters that define the grid parameter `̀`l`` as a dict with keys ``{"a", "b", "plastic"}``. See below for more detail.
        fields_names: the names of the scalar grid-shaped fields
        equation_nl: the function that performs the nonlinear update step. Returns the time derivative of the fields
        equation_l: implicit visocsity update step, called after the RK step has ended. Returns the new fields and their new time derivative
        k_min: minimum k of the grid. Default is defined by :class:`pyloggrid.LogGrid.Grid.Grid`.
        k0: whether there's a k0 mode
        simu_params: physical quantities relevant to the simulation, fixed for the whole simulation.
        n_threads: Number of threads to use, default is max/2 (not recommended, run benchmarking for better results. If running batch simulations, 1 is optimal.)

    In `pyloggrid`, simulating equations is done through the :class:`Solver` class.

    Initializing the simulation
    ###########################

    A `Solver` object is initialized with parameters of the simulation:

    * Geometric parameters (``D, l_params, k_min, k0``)
        * ``l_params`` is a ``dict`` that describes the scaling parameter ``l`` of the grid. If the "plastic" key is set to ``True`` (``l_params={"plastic": True, ...}``) then ``l~=1.3`` (the plastic number). If ``plastic`` is not ``True``, then keys ``a, b`` must be specified and have integer values, in which case ``l`` is the solution to ``l**b-l**a=1``. In particular if ``l_params={"a": 1, "b": 2}`` then ``l~=1.6`` (the golden number). To get the special case ``l=2``, use ``l_params={"a":None, "b": None}``.

        .. important:: Choosing either ``"plastic": True``, ``"a": int, "b": int`` or ``"a": None, "b": None`` doesn't just change the scaling of the grid, but also the number of interactions per point for the convolution (listed here in decreasing order), which affects both the physics and the required computing power.

        * ``k0`` is a bool that indicates whether the axes :math:`k_i=0` are included. Enabling this comes at a significant performance cost, and greatly modifies the interaction between grid points.

        * ``k_min`` is the minimum absolute value of a wavevector *on each axis*, excluding zeros. This is **not** the minimum magnitude of a wavevector on the grid, which would be :math:`\\sqrt{D}k_{min}` in :math:`D` dimensions.

    * Equation parameters (``field_names``, ``equation_nl``, ``equation_l``, ``simu_params``)
        * For stability and performance reasons, the equation to solve is divided into two parts: one that is proportional to the scalar input ``equation_l``, and the rest ``equation_nl``. If your equation is :math:`\\partial_t u=A(u)−b\\cdot u`, then the linear part is :math:`−b`, and the nonlinear part is :math:`A(t, u)`.
        * ``equation_l, equation_nl`` expect functions that take as arguments ``t: float, grid: Grid, simu_params: dict`` and return a dict ``{"field_name": d/dt_field_name, ...}``.

        .. warning:: Note that the linear part must be time-independent.

        * ``simu_params`` are arbitrary (serializable) constants used in the simulation. The ``fields`` named by ``fields_names``, on the other hand, are grid-shaped complex scalars., i.e. "the fields being simulated".

    Running the simulation
    ######################

    Simulations are ran by calling the :func:`solve` method.

    .. warning:: Although in theory you can use the ``solve`` method several times on a single ``Solver`` object, this isn't a common use case. Do so at your own risks, but please report unexpected behavior to us.

    """

    def __init__(
        self,
        D: int,
        l_params: dict[str, Union[float, bool]],
        fields_names: list[str],
        equation_nl: Callable[[float, Grid, dict], dict],
        equation_l: Callable[[float, Grid, dict], dict[str, np.ndarray]],
        k_min: float = None,
        k0: bool = False,
        simu_params: dict = None,
        n_threads: int = None,
    ):
        self.equation_nl = equation_nl
        self.equation_l = equation_l
        self.simu_params = simu_params
        self.end_simulation = None

        # Initialize internal counters
        self.step = 0  # current saved step. Equal to ode_step if we save every step
        self.ode_step = 0  # current simulation step (+1 for every true [= not rejected] step). Not equal to number of calls to equation_nonlinear, but equal to number of calls to equation_linear.
        self.time_tracker = TimeTracker()
        self.time_tracker.start_timer("total")
        self.grid = Grid(D=D, l_params=l_params, N_points=1, k_min=k_min, fields_name=fields_names, n_threads=n_threads, k0=k0)
        self.grid.time_tracker = self.time_tracker

    def _load_parameters(self, path: str) -> tuple[float, float]:
        """Loads simulation parameters from settings.json.
        Overwrites instance variables

        Args:
            path: directory to load the settings from

        Returns:
            (last used time, last used timstep)
        """
        settings = read_json_settings(path)

        # Overwrite instance variables
        D = settings["D"]
        self.simu_params = settings["simu_params"]
        self.end_simulation = settings["end_simulation"]
        l_params = settings["l_params"]
        fields_name = settings["fields_name"]

        # Load last data step & overwrite relevant instance variables
        step = settings["N_steps"]
        ode_step = settings["N_ode_steps"]
        step_data = load_step(path=path, step=step, fields_names=fields_name)
        k0 = step_data["k0"]
        self.step = step
        self.ode_step = ode_step
        self.time_tracker = TimeTracker({k[7:]: v for k, v in step_data.items() if k[:7] == "ttrack_"})
        self.time_tracker.start_timer("total")
        N_points = next(iter(step_data["fields"].values())).shape[0] - (1 if k0 else 0)
        self.grid = Grid(
            D=D, l_params=l_params, N_points=N_points, k_min=step_data["k_min"], fields_name=fields_name, n_threads=self.grid.maths.n_threads, k0=k0
        )
        self.grid.fields = step_data["fields"]
        self.grid.time_tracker = self.time_tracker

        return float(step_data["t"]), float(step_data["dt"])

    def _save_step_all(self, t: float, dt: float, path: str) -> None:
        """Save the current step. Saves both fields and settings.

        Args:
            t: current time
            dt: timestep
            path: folder to save to. Will override any existing file.
        """
        self.step += 1
        logger.info(f"Saving real step {self.step}, t={t:8}")

        save_step(path, step=self.step, grid=self.grid, t=t, dt=dt, time_tracker=self.time_tracker)
        update_json_settings(path, {"N_steps": self.step, "N_ode_steps": self.ode_step}, update=True)

    def solve(
        self,
        save_path: str,
        initial_conditions: str | Callable[[dict[str, np.ndarray], Grid, dict], dict[str, np.ndarray]] | tuple | list,
        solver_params: dict = None,
        init_t: float = 0,
        end_simulation=None,
        save_one_in: float = 1,
        update_gridsize_cb: Optional[Callable[[Grid], Optional[int]]] = None,
        dt_params: dict[str, Optional[float]] = None,
        solver: typing.Literal["ViscDopri", "ETD4RK", "ETD35"] = "ETD35",
    ) -> None:
        """Solve the solver's equation with the provided numerical parameters.

        Args:
            save_path: where to save the fields and simulation parameters. Will backup if exists, will create otherwise
            initial_conditions: if str: either ``"loadfromsave"`` to resume simulation, [legacy: or the path of the ``.npz`` save file whence to initialize the fields].
                If tuple ``(name, step)``: load step ``step`` from .h5 ``name``.
                Otherwise: function that returns the initial fields.
            solver_params: params forwarded to the solver
            init_t: initial simulation time
            end_simulation: dict of thresholds to end the simulation: ``{t: max physical time, elapsed_time: max real time spent computing, step: max save step}``
            save_one_in: save one step to file every X ``ode_step``
            update_gridsize_cb: optional callback to change the grid size after a step. Returns ``None`` if no change is to be made, returns the new grid size otherwise.
            dt_params: ``{dt0: initial timestep, dtmin, dtmax}``
            solver: ``"ETD35"`` for :class:`ETD35` (default), ``"ETD4RK"`` for :class:`ETD4RK`, ``"ViscDopri"`` for :class:`ViscDopri`

        Details on a few parameters
        ###########################

        * If ``save_path="<path>/<foldername>"`` already exists on disk, the existing directory will be renamed to ``<path>/bk_<foldername>_<TIMESTAMP>``. Otherwise, it will be automatically created.
        * If ``initial_conditions="loadfromsave"`` then settings and initial conditions will be loaded from the last step of the existing simulation in ``save_path``. This will override the following arguments: ``end_simulation, l_params, D, fields_name, simu_params, k_min, k0``.
        * To load from a specific step of a .h5 simulation file, set ``loadfromsave=(<filepath>:str, <step>:int)``.
        * Otherwise, ``initial_conditions`` must be a function that takes (fields, grid, simu_params) and returns the initial fields.
        * ``update_gridsize_cb`` is an optional function to dynamically change the grid size. It takes (grid) and returns either None (no change) or a new grid size.
        """
        if solver_params is None:
            solver_params = {}
        if end_simulation is None:
            end_simulation = {}
        end_simulation = {
            "t": end_simulation["t"] if "t" in end_simulation else None,
            "elapsed_time": end_simulation["elapsed_time"] if "elapsed_time" in end_simulation else None,
            "step": end_simulation["step"] if "step" in end_simulation else None,
            "ode_step": end_simulation["ode_step"] if "ode_step" in end_simulation else None,
        }
        self.end_simulation = end_simulation

        # Initial conditions (IC)
        if callable(initial_conditions):
            fields = initial_conditions(self.grid.fields, self.grid, self.simu_params)
            self.grid = self.grid.to_new_size(fields)
        elif initial_conditions == "loadfromsave":
            params = {"end_simulation": end_simulation, "solver_params": solver_params, "simu_params": self.simu_params}  # override previous parameters
            update_json_settings(save_path, params, update=True)

            init_t, dt = self._load_parameters(save_path)
            if dt > 0:  # dt=0 if only the initial step is saved
                if dt_params is None:
                    dt_params = {}
                dt_params["dt0"] = dt
            logger.info(f"Starting again from t={init_t}, dt={dt:.2e}, step={self.step}, grid size N={self.grid.N_points}")
        else:
            if isinstance(initial_conditions, (tuple, list)):
                logger.info(f"Loading IC from: {initial_conditions[0]} step {initial_conditions[1]}")
                step_data = load_step(self.grid.fields_names, filepath=initial_conditions[0], step=initial_conditions[1])
            elif isinstance(initial_conditions, str):  # load from a given .npz step, legacy method
                logger.info(f"Loading IC from: {initial_conditions}")
                step_data = load_step(filepath=initial_conditions, fields_names=self.grid.fields_names)
            else:
                raise ValueError("Wrong initial_conditions provided")

            self.grid = Grid(
                D=self.grid.D,
                l_params=self.grid.l_params,
                N_points=step_data["N_points"],
                k_min=step_data["k_min"],
                fields_name=self.grid.fields_names,
                n_threads=self.grid.maths.n_threads,
                k0=step_data["k0"],
            )
            self.grid.fields = step_data["fields"]
            self.grid.time_tracker = self.time_tracker

        for field_name in self.grid.fields:  # make sure someone didn't forget to cast as complex in init conditions...
            self.grid.fields[field_name] = self.grid.fields[field_name].astype(complex)

        def equation_nl_convertdim(t: float, y: np.ndarray) -> np.ndarray:
            """ "Man-in-the-middle" update step that converts N-D <-> 1D.

            Args:
                t: time
                y: 1D DOPRI array

            Returns:
                the 1D result of the update step
            """

            # 1D -> N-D
            new_fields = D1_to_fields(y, self.grid.fields_names, self.grid.shape)
            self.grid.fields = new_fields  # sets grid fields to current values

            # N-D -> 1D
            return fields_to_1D(self.grid.maths.enforce_grid_symmetry_dict(self.equation_nl(t, self.grid, self.simu_params)), self.grid.fields_names)

        def save_step(t: float, dt: float, y: np.ndarray) -> bool:
            """Save the current step and check if the simulation should end
            Also enforces symmetry if needed

            Args:
                t: time
                dt: time step used
                y: 1D DOPRI array

            Returns:
                whether the sim should end
            """
            newy = D1_to_fields(y, self.grid.fields_names, self.grid.shape)

            self.grid.fields = newy  # sets grid fields to current values
            self.grid.enforce_grid_symmetry()

            # update solver parameters
            self.ode_step += 1
            self.time_tracker.tick_timer("total")
            logger.debug(f"Solver: ode step {self.ode_step}, t={t}, dt={dt}, elapsed time: {datetime.timedelta(seconds=self.time_tracker.get('total'))}")

            # Check whether to stop the simulation
            stopsim = False
            sst, sset, sss, ssos = self.end_simulation["t"], self.end_simulation["elapsed_time"], self.end_simulation["step"], self.end_simulation["ode_step"]
            if sst is not None and t >= sst:
                logger.info(f"Reached time: {sst}")
                stopsim = True
            elif sset is not None and self.time_tracker.get("total") >= sset:
                logger.info(f"Reached elapsed time: {sset}")
                stopsim = True
            elif sss is not None and self.step >= sss:
                logger.info(f"Reached saved step: {sss}")
                stopsim = True
            elif ssos is not None and self.ode_step >= ssos:
                logger.info(f"Reached ode step: {ssos}")
                stopsim = True

            # Check for grid size update
            with self.time_tracker("update_gridsize"):
                new_N_points = None
                if update_gridsize_cb is not None:
                    new_N_points = update_gridsize_cb(self.grid)
            if new_N_points is not None and not stopsim:
                self.time_tracker.start_timer("update_gridsize")
                logger.info(f"Updating the grid: old N = {self.grid.N_points}, new N = {new_N_points}")
                self.grid = self.grid.to_new_size_empty(new_N_points).load_fields(self.grid.fields)
                self._save_step_all(t, dt, save_path)

                logger.info("Starting new solver on updated grid")
                self.time_tracker.end_timer("update_gridsize")
                self.solve(
                    solver_params=solver_params,
                    initial_conditions="loadfromsave",
                    init_t=init_t,
                    save_path=save_path,
                    save_one_in=save_one_in,
                    end_simulation=end_simulation,
                    dt_params=dt_params,
                    update_gridsize_cb=update_gridsize_cb,
                    solver=solver,
                )
                raise SolverInterruptedError()

            # Save data
            with self.time_tracker("disk"):
                if stopsim or self.ode_step % save_one_in == 0:
                    self._save_step_all(t, dt, save_path)

            return stopsim

        def equation_l_convertdim(t: float, y: np.ndarray) -> np.ndarray:
            """Convert the linear array 1D <-> ND

            Args:
                t: time
                y: 1D DOPRI array

            Returns:
                1D new y, 1D new dy/dt
            """

            new_fields = D1_to_fields(y, self.grid.fields_names, self.grid.shape)
            self.grid.fields = new_fields
            visc = self.equation_l(t, self.grid, self.simu_params)

            return fields_to_1D(visc, self.grid.fields_names)

        # load / initialize parameters & save paths
        if initial_conditions != "loadfromsave":
            # Check for existing results
            if os.path.isdir(save_path):
                # Rename the folder just in case the user forgot to save it
                bk_name = f"{save_path}_bk_{int(time.time())}"
                os.rename(save_path, bk_name)
                logger.info(f"Save dir was not empty, renamed to {bk_name}")
                # make new folder
                os.makedirs(save_path)
            else:
                logger.info(f"Folder {save_path} does not exist, creating a new one")
                os.makedirs(save_path)

            # save current simulation file to keep a trace of what we ran
            orig_file = sys.modules["__main__"].__file__
            shutil.copyfile(orig_file, os.path.join(save_path, "source.py"))

            # initialize and save solver parameters
            params = {
                "init_t": init_t,
                "end_simulation": end_simulation,
                "N_steps": 0,
                "N_ode_steps": 0,
                "l_params": self.grid.l_params,
                "D": self.grid.D,
                "solver_params": solver_params,
                "fields_name": self.grid.fields_names,
                "simu_params": self.simu_params,
            }
            update_json_settings(save_path, params)
            init_t = params["init_t"]

        # Start solver
        logger.info("*** Starting solver ***")

        match solver:
            case "ViscDopri":
                ode = ViscDopri(
                    equation_nl=equation_nl_convertdim,
                    equation_l=equation_l_convertdim,
                    init_t=init_t,
                    solver_params=solver_params,
                    y0=fields_to_1D(self.grid.fields, self.grid.fields_names),
                    dt_params=dt_params if dt_params is not None else {},
                    save_step=save_step,
                )
            case "ETD4RK":
                ode = ETD4RK(
                    equation_nl=equation_nl_convertdim,
                    equation_l=equation_l_convertdim,
                    init_t=init_t,
                    y0=fields_to_1D(self.grid.fields, self.grid.fields_names),
                    dt_params=dt_params if dt_params is not None else {},
                    save_step=save_step,
                )
            case "ETD35":
                ode = ETD35(
                    equation_nl=equation_nl_convertdim,
                    equation_l=equation_l_convertdim,
                    init_t=init_t,
                    solver_params=solver_params,
                    y0=fields_to_1D(self.grid.fields, self.grid.fields_names),
                    dt_params=dt_params if dt_params is not None else {},
                    save_step=save_step,
                )
            case _:
                raise ValueError(f"Unknown solver provided: {solver}")
        if initial_conditions != "loadfromsave":
            self._save_step_all(init_t, 0, save_path)  # save first step
        with contextlib.suppress(SolverInterruptedError):  # Used to restart with a different grid
            ode.solve()
            logger.info(f"Finished sim. at t={ode.t}, success: {ode.status}, ode_steps: {self.ode_step}, saved steps:{self.step}, elapsed: {self.time_tracker}")



"""used to package the source and handle compilation"""

import os
import pathlib
import sys

from setuptools import Extension
from setuptools.command.build_ext import build_ext as build_ext_orig


class MakefileExtension(Extension):
    def __init__(self, name):
        # don't invoke the original build_ext for this special extension
        super().__init__(name, sources=[])


class build_ext(build_ext_orig):
    def run(self):
        for ext in self.extensions:
            self.build_makefile(ext)

    def build_makefile(self, ext):
        cwd = pathlib.Path().absolute()
        # this seems necesssary to keep the files installed ?
        extdir = pathlib.Path(self.build_lib) / ext.name
        makefile_dir = str(extdir.parent.absolute())

        try:
            os.chdir(makefile_dir)
            if sys.platform == "win32":
                self.spawn(["make", "clean", "-f", "Makefile.windows"])
                self.spawn(["make", "-f", "Makefile.windows"])
            else:
                self.spawn(["make", "clean"])
                self.spawn(["make"])

            print("Module compilation completed successfully.")
        except Exception as e:
            print(f"Error compiling modules: {str(e)}")
        finally:
            # Change back to the original working directory
            os.chdir(str(cwd))


def build(setup_kwargs: dict):
    setup_kwargs |= dict(
        cmdclass={"build_ext": build_ext},
        ext_modules=[MakefileExtension("pyloggrid/LogGrid/convolver")],
    )



"""Utilities for fast plotting"""

from __future__ import annotations

from pyloggrid.Libs.misc import rightpad_array_2D
from pyloggrid.Libs.singlethread_numpy import np

np.zeros(0)  # Do not remove - used to force singlethread_numpy to be imported before the other libraries when refactoring
import typing

import imageio.v2 as imageio
import matplotlib
import scienceplots  # needs to be imported, even if not called directly
from matplotlib import cycler
from matplotlib import pyplot as plt
from matplotlib.figure import Figure
from matplotlib.widgets import Slider

# noinspection PyStatementEffect
scienceplots.stylesheets  # prevent removal of import

if typing.TYPE_CHECKING:
    from typing import Any, Callable

    # noinspection PyPackageRequirements
    from cycler import Cycler
    from matplotlib.colors import LinearSegmentedColormap

    from pyloggrid.LogGrid.Grid import Grid


# Unified formatting
def initFormat(size: float = 1, txtsize: float = None, rcParams: dict = None) -> None:
    """Sets up pyplot formatting for scientific publications.

    Args:
        size: the figure's size relative to the default one
        txtsize: scaling for the figure's fonts, if not set defaults to ``size``
        rcParams: parameters to append to :attr:`matplotlib.rcParams`
    """

    if rcParams is None:
        rcParams = {}

    def force_add_cycler(cyclers: list["Cycler"]) -> "Cycler":
        """add two cyclers, whatever their length"""
        if len(cyclers) == 2:
            cyc_a, cyc_b = cyclers
            try:
                return cyc_a + cyc_b
            except ValueError:
                return cyc_a * len(cyc_b) + cyc_b * len(cyc_a)
        else:
            return force_add_cycler([cyclers[0], force_add_cycler(cyclers[1:])])

    if txtsize is None:
        txtsize = size

    plt.style.use(["science", "grid"])
    plt_params = {
        "figure.figsize": (8 * size, 6 * size),
        "font.size": 22 * txtsize,
        "axes.labelsize": 20 * txtsize,
        "axes.titlesize": 20 * txtsize,
        "legend.fontsize": 16 * txtsize,
        "xtick.labelsize": 16 * txtsize,
        "ytick.labelsize": 16 * txtsize,
        "lines.markersize": 6 * size,
        "lines.linewidth": 2 * size,
        "font.family": "serif",
        "font.serif": ["Times New Roman"],
        "text.usetex": False,
        "xtick.direction": "out",
        "ytick.direction": "out",
        "xtick.top": False,
        "ytick.right": False,
        "xtick.minor.visible": False,
        "ytick.minor.visible": False,
        "legend.framealpha": 0.7,
        "legend.fancybox": False,
        "axes.prop_cycle": force_add_cycler(
            [
                cycler("color", ["#006BA4", "#FF800E", "#ABABAB", "#595959", "#5F9ED1", "#C85200", "#898989", "#A2C8EC", "#FFBC79", "#CFCFCF"]),
                cycler("linestyle", ["-", "--", "-.", ":"]),
                cycler("marker", ["o", "d", "v", "^", "<", ">", "1", "2", "3", "4", "8", "s", "p", "P", "*", "h", "H", "+", "x", "X", "D", "|", "_"]),
            ]
        ),
        "image.cmap": "cividis",
    }
    matplotlib.rcParams.update(plt_params | rcParams)


# Common
def pltshowm(full: bool = True, save: str = None, legend: bool = True, compact: bool = True, tight: bool = False) -> None:
    """Display the pyplot figure maximized.

    Args:
        legend: if ``False``, do not add a legend
        save: if specified, save path for the figure
        full: if ``False``, doe not go fullscreen
        compact: if ``True``, use ``plt.tight_layout``
        tight: if ``True``, autoscale with tight boundaries
    """
    if full:
        man = plt.get_current_fig_manager()
        try:
            man.frame.Maximize(True)
        except AttributeError:  # wrong backend, try QT
            try:
                man.window.showMaximized()
            except AttributeError:  # probably running a backend that can only save the image, not display it
                matplotlib.pyplot.gcf().set_size_inches(19.2, 10.8)
    if legend:
        plt.legend()
    if tight:
        plt.autoscale(tight=True)
    if compact:
        plt.tight_layout()
    if save:
        plt.savefig(save, dpi=300, bbox_inches="tight")
    plt.show()


def labels(xlabel: str = "", ylabel: str = "", title: str = "", ax: Any = None) -> None:
    """Add labels and title to an axis.

    Args:
        xlabel
        ylabel
        title
        ax: if specified, the axis on which to add the labels; defaults to the current axis
    """
    ax = ax or plt.gca()
    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)
    ax.set_title(title)


def scatter(ax=None, *args, **kwargs) -> Any:
    """Plot a scatter plot using ``plot()`` instead of ``scatter()``.

    (not handled the same way, e.g. zindex)

    Args:
        ax: the axis to use, defaults to
    """
    ax = ax or plt.gca()
    if "linestyle" not in kwargs:
        kwargs["linestyle"] = ""
    return ax.plot(*args, **kwargs)


# Axes
def plot2axes() -> tuple[Figure, Any, Any]:
    """Create a plot with left and right axis.

    Axes are set to `Any` since the real class is hidden in ``mpl.axes._subplots``.

    Returns:
        ``figure, ax1, ax2``
    """
    fig, ax1 = plt.subplots()
    ax2 = ax1.twinx()
    return fig, ax1, ax2


def plot2axesX() -> tuple[Figure, Any, Any]:
    """plot with bottom and top axis.

    Axes are set to `Any` since the real class is hidden in ``mpl.axes._subplots``.
    """
    fig, ax1 = plt.subplots()
    ax2 = ax1.twiny()
    return fig, ax1, ax2


# Sliders
def _save_animation_from_slider(slider: Slider, save: str, temp_dir: str, N: int = 100) -> None:
    """Save a matplotlib graph with a slider as an animation.

    Args:
        slider: the slider object
        save: the output file name
        temp_dir: where to store the intermediate images
        N: number of frames
    """

    maxw, maxh = 0, 0
    for n in np.linspace(slider.valmin, slider.valmax, N):
        slider.set_val(n)
        plt.savefig(f"{temp_dir}/{n}.png")
        img = imageio.imread(f"{temp_dir}/{n}.png")
        w, h = img.shape[:-1]
        maxw, maxh = max(w, maxw), max(h, maxh)

    # noinspection PyTypeChecker
    with imageio.get_writer(save, format="FFMPEG", mode="I", fps=10, output_params=["-vf", "scale=500:500", "-sws_flags", "neighbor"]) as writer:
        for n in np.linspace(slider.valmin, slider.valmax, N):
            img = imageio.imread(f"{temp_dir}/{n}.png")
            w, h = img.shape[:-1]
            img = rightpad_array_2D(img, maxw - w, maxh - h)
            writer.append_data(img.astype(np.uint8))


def dec_enable_slider_save(f: Callable) -> Callable:
    """Decorator to save an animation from a slider graph.

    This adds two arguments to the decorated function, ``save`` and ``temp_path``, which if supplied are forwarded to :func:`_save_animation_from_slider`.
    """

    def wrapper(*args, save: str = None, temp_path: str = None, **kwargs):
        """:param save if not None, save an animation of the slider as mp4 to a file named <save>
        :param temp_path if save is not None, where to store generated images
        """
        res = f(*args, **kwargs)
        slider = res[0]
        if save:
            assert temp_path is not None, "You must provide <temp_path>"
            _save_animation_from_slider(slider, save, temp_path)
        return res

    return wrapper


@dec_enable_slider_save
def interactive_spectrum(ts: np.ndarray, ks: np.ndarray, spectra: dict[str, np.ndarray], quantities: dict[str, np.ndarray]) -> tuple[Slider, Any, Any]:
    r"""An interactive spectrum + arbitrary time-dependant quantities.

    Warning:
        As for all interactive functions, the returned val must be assigned, ex via ``_ = my_interactive_function(...)``

    LHS: spectrum vs k
    RHS: quantity vs t
    current time is changed via a slider

    Args:
        ts: time array
        ks: ks array for each time
        spectra: dict label->array
        quantities: dict label->quantity
    """

    fig, axs = plt.subplot_mosaic("AB")
    ax1, ax2 = axs["A"], axs["B"]

    ax1.set_xscale("log")
    ax1.set_yscale("log")
    padfact = 2
    ax1.set_xlim(min(np.min(i) for i in ks) / padfact, max(np.max(i) for i in ks) * padfact)
    ax1.set_ylim(1e-200, max(np.max(i) for i in list(spectra.values())[0]) * padfact * 1e5)
    ax1.set_xlabel("$k$")
    ax1.set_ylabel("spectra")

    # noinspection PyMissingOrEmptyDocstring
    def update_graph(_=None):
        i = int(n_slider.val - 1)
        t = ts[i]
        vl.set_xdata([t, t])
        for k, l in spectrum_lines.items():
            l.set_data(ks[i], spectra[k][i])

    # Sliders & Buttons
    fig.subplots_adjust(bottom=0.2)
    n_slider = Slider(ax=fig.add_axes((0.25, 0.1, 0.65, 0.03)), label="Step", valfmt="%d", valmin=1, valmax=ts.size, valinit=1, valstep=1)
    n_slider.on_changed(update_graph)

    spectrum_lines = {}
    for k in spectra:
        spectrum_lines[k] = scatter(ax1, [0], [1], label=k)[0]
    ax1.legend()

    for k, v in quantities.items():
        scatter(ax2, ts, v, label=k)
    ax2.legend()
    vl = ax2.axvline(ts[0], ls="-", color="r", lw=1, zorder=10)
    ax2.set_xlabel("$t$")

    update_graph()
    return n_slider, ax1, ax2


@dec_enable_slider_save
def interactive_3D_logplot_by_z(X: np.ndarray, Y: np.ndarray, V: np.ndarray, Z: np.ndarray = None) -> tuple[Slider, Any, Any]:
    """An interactive 3D logplot, sliced by ``Z`` if given.

    Warning:
        As for all interactive functions, the returned val must be assigned, ex via ``_ = my_interactive_function(...)``

    You can either provide 3D arrays (X, Y, V, Z[:,:,0]) or 2D arrays (X, Y, V).

    Args:
        X
        Y
        V: values
        Z: optional 1D Z axis
    """
    import matplotlib.ticker as mticker
    from matplotlib.widgets import Button, Slider

    # Set the loglog scale for the x, y, and z axes
    # noinspection PyMissingOrEmptyDocstring
    def asinh(data, a0: 1):
        return np.arcsinh(data / a0) * a0

    # noinspection PyMissingOrEmptyDocstring
    def sinh(data, a0: 1):
        return np.sinh(data / a0) * a0

    ## animation code
    factk, fact0, z_mode = None, None, Z is None

    # noinspection PyMissingOrEmptyDocstring
    def get_toplot(doall=False):
        if doall:
            toplot_ = V
        elif z_mode:
            toplot_ = V[int(n_slider.val) - 1]
        else:
            toplot_ = V[int(n_slider.val) - 1, :, :, int(z_slider.val)]
        return toplot_

    # noinspection PyMissingOrEmptyDocstring
    def update_graph(_=None):
        nonlocal factk, fact0
        n_z = int(z_slider.val) if Z is not None else None
        toplot_ = get_toplot()
        X_, Y_ = (X, Y) if z_mode else (X[:, :, n_z], Y[:, :, n_z])

        graph._offsets3d = np.array([asinh(X_.flatten(), factk), asinh(Y_.flatten(), factk), asinh(toplot_.real.flatten(), fact0)])  # position
        if z_mode and Z is not None:
            sel = np.zeros_like(toplot_.real)
            sel[:] = -1
            sel[:, :, n_z] = 1
            graph.set_array(sel.flatten())
            sizes = np.zeros_like(toplot_.real)
            sizes[:] = 10
            sizes[:, :, n_z] = 30
            graph.set_sizes(sizes.flatten())
        else:
            graph.set_array(toplot_.imag.flatten() / np.abs(toplot_.flatten()))
            graph.set_sizes(np.ones_like(toplot_.real.flatten()) * 10)

        if Z is not None:
            z_slider_text.set_text(f"z={Z[n_z]:.2e}")

        fig.canvas.draw_idle()

    def rescale(_=None, doall=False):
        """rescale the graph"""
        nonlocal factk, fact0
        toplot_ = get_toplot(doall)

        factk = np.min(np.abs(X)) / 10
        fact0 = max(1e-250, np.min(np.abs(toplot_.real[toplot_.real != 0])) / 10)

        ax.set_xscale("asinh", linear_width=factk)  # just for the lines, before setting the ticks
        ax.set_yscale("asinh", linear_width=factk)
        ax.set_zscale("asinh", linear_width=fact0)

        ax.set_xlim(np.min(asinh(X.flatten(), factk)), np.max(asinh(X.flatten(), factk)))
        ax.set_ylim(np.min(asinh(Y.flatten(), factk)), np.max(asinh(Y.flatten(), factk)))
        ax.set_zlim(np.min(asinh(toplot_.real.flatten(), fact0)), np.max(asinh(toplot_.real.flatten(), fact0)))
        ax.xaxis.set_major_formatter(mticker.FuncFormatter(lambda val, _: f"{sinh(val, factk):.2e}"))
        ax.xaxis.set_major_locator(mticker.MaxNLocator(nbins=5))
        ax.yaxis.set_major_formatter(mticker.FuncFormatter(lambda val, _: f"{sinh(val, factk):.2e}"))
        ax.yaxis.set_major_locator(mticker.MaxNLocator(nbins=5))
        ax.zaxis.set_major_formatter(mticker.FuncFormatter(lambda val, _: f"{sinh(val, fact0):.2e}"))
        ax.zaxis.set_major_locator(mticker.MaxNLocator(nbins=5))
        ax.zaxis.set_major_locator(mticker.AsinhLocator(linear_width=fact0, subs=(2, 5), symthresh=0.0, numticks=5))

        update_graph()

    # noinspection PyMissingOrEmptyDocstring
    def toggle_z(_=None):
        nonlocal z_mode
        z_mode = not z_mode
        update_graph()

    fig = plt.figure()
    ax = fig.add_subplot(111, projection="3d")

    ## Sliders & Buttons
    # fig.subplots_adjust(bottom=-0.1)
    n_slider = Slider(
        ax=fig.add_axes((0.25, 0.1, 0.65, 0.03)),
        label="Step",
        valfmt="%d",
        valmin=1,
        valmax=V.shape[0],
        valinit=1,
        valstep=1,
    )
    n_slider.on_changed(update_graph)
    z_slider, btn_toggle_z = None, None
    if Z is not None:
        z_slider_ax = fig.add_axes((0.1, 0.25, 0.0225, 0.63))
        z_slider = Slider(ax=z_slider_ax, label="n_z", valfmt="%d", valmin=0, valmax=X.shape[2] - 1, valinit=X.shape[2] // 2, valstep=1, orientation="vertical")
        z_slider_text = z_slider_ax.text(0.5, 0.5, "")
        z_slider.on_changed(update_graph)

    btn_rescale = Button(fig.add_axes((0.8, 0.025, 0.1, 0.04)), "Rescale", hovercolor="0.975")
    btn_rescale.on_clicked(rescale)
    if Z is not None:
        btn_toggle_z = Button(fig.add_axes((0.6, 0.025, 0.1, 0.04)), "Toggle z mode", hovercolor="0.975")
        btn_toggle_z.on_clicked(toggle_z)

    graph = ax.scatter([], [], [], label=r"$\Re ux$", alpha=0.5)
    ax.set_xlabel("$X$")
    ax.set_ylabel("$Y$")
    rescale(doall=True)
    graph.set_cmap("coolwarm")

    return n_slider, ax, (graph, z_slider, btn_rescale, btn_toggle_z)


@dec_enable_slider_save
def interactive_grid_imshow(
    grid: Grid,
    update_data: Callable[[Any, float], np.ndarray],
    slider_params: tuple[float | None, float | None, float | None, str | None, str | None] = (0, 1, None, "", None),
) -> tuple[Slider, Any]:
    """Imshow with slider, with axes formatted by the grid.

    Warning:
        As for all interactive functions, the returned val must be assigned, ex via ``_ = my_interactive_function(...)``

    Args:
        grid
        update_data: a function that takes the plotting ax and the slider's value, and returns an array to plot of size (N, N)
        slider_params: (vmin, vmax, vstep, title, vformat), all optional
    """
    fig, ax = plt.subplots()
    N = grid.N_points

    # noinspection PyMissingOrEmptyDocstring
    def update_graph(_=None):
        im.set_array(update_data(ax, n_slider.val))
        im.autoscale()

    im = ax.imshow(np.zeros((N, N)), origin="lower", interpolation="none")
    ticks = np.linspace(0, N, 5)
    ax.set_xlabel("$k_2$")
    ax.set_ylabel("$k_1$")
    ax.set_xticks(ticks)
    ax.set_yticks(ticks)
    ax.set_xticklabels([f"$10^{{{int(np.log10(i))}}}$" for i in [grid.k_min * grid.l**j for j in ticks]])
    ax.set_yticklabels([f"$10^{{{int(np.log10(i))}}}$" for i in [grid.k_min * grid.l**j for j in ticks]])
    plt.colorbar(im)

    ## Sliders & Buttons
    fig.subplots_adjust(bottom=0.2)
    vmin, vmax, vstep, slidertitle, vformat = slider_params
    n_slider = Slider(ax=fig.add_axes([0.25, 0.1, 0.65, 0.03]), label=slidertitle, valfmt=vformat, valmin=vmin, valmax=vmax, valinit=vmin, valstep=vstep)
    n_slider.on_changed(update_graph)

    update_graph()

    return n_slider, ax


@dec_enable_slider_save
def interactive_grid_3Dslice(
    grid: Grid,
    update_data: Callable[[Any, float], np.ndarray],
    slider_params: tuple[float | None, float | None, float | None, str | None, str | None] = (0, 1, None, "", None),
) -> tuple[Slider, Any]:
    """Interactive 3D slice of a grid field.

    Warning:
        As for all interactive functions, the returned val must be assigned, ex via ``_ = my_interactive_function(...)``

    Args:
        grid
        update_data: a function that takes the plotting ax and the slider's value, and returns a 3D array to plot
        slider_params: (vmin, vmax, vstep, title, vformat), all optional
    """
    fig = plt.figure()
    ax = fig.add_subplot(projection="3d")
    cmap = matplotlib.colormaps.get("gist_rainbow")
    N = grid.N_points

    # noinspection PyMissingOrEmptyDocstring
    def update_graph(_=None):
        data = update_data(ax, n_slider.val)

        ax.cla()
        kxy = range(N)
        kz = range(N)
        KXY, KZ = np.meshgrid(kxy, kz)
        # primary plane
        data_plane1 = np.array([[data[i, N + i, N + j] for j in kz] for i in kxy])
        data_plane1 = data_plane1 / np.max(data_plane1)
        ax.plot_surface(KXY, KXY, KZ, facecolors=cmap(data_plane1), alpha=0.95, linewidth=0)
        # secondary plane
        data_plane2 = np.array([[data[j, N + i, N + i] for j in kz] for i in kxy])
        data_plane2 = data_plane2 / np.max(data_plane2)
        ax.plot_surface(KZ, KXY, KXY, facecolors=cmap(data_plane2), alpha=0.55, linewidth=0)

    ax.set_xlabel("$k_x$")
    ax.set_ylabel("$k_y$")
    ax.set_zlabel("$k_z$")
    m = matplotlib.cm.ScalarMappable(cmap=cmap)
    m.set_array([])
    plt.colorbar(m, ax=ax)

    ## Sliders & Buttons
    fig.subplots_adjust(bottom=0.2)
    vmin, vmax, vstep, slidertitle, vformat = slider_params
    n_slider = Slider(ax=fig.add_axes((0.25, 0.1, 0.65, 0.03)), label=slidertitle, valfmt=vformat, valmin=vmin, valmax=vmax, valinit=vmin, valstep=vstep)
    n_slider.on_changed(update_graph)

    update_graph()

    return n_slider, ax


@dec_enable_slider_save
def interactive_3D_logplot_positive(
    update_data: Callable[[Any, float], np.ndarray],
    slider_params: tuple[float | None, float | None, float | None, str | None, str | None] = (0, 1, None, "", None),
    threshold_fact: float = 5,
) -> tuple[Slider, Any]:
    """3D logplot of scalar field on kx,ky,kz>0.

    Warning:
        As for all interactive functions, the returned val must be assigned, ex via ``_ = my_interactive_function(...)``

    Args:
        update_data: a function that takes the plotting ax and the slider's value, and returns a 3D array to plot
        slider_params: (vmin, vmax, vstep, title, vformat), all optional
        threshold_fact: which points to hide, relative to max(data)
    """
    fig = plt.figure()
    ax = fig.add_subplot(111, projection="3d")
    ax.set(xlabel="$k_x$")
    ax.set(ylabel="$k_y$")
    ax.set(zlabel="$k_z$")
    graph = ax.scatter([], [], [], s=50, cmap="bwr", vmin=0, vmax=1, alpha=0.5)
    graph.set_cmap("bwr")

    # noinspection PyMissingOrEmptyDocstring
    def update_graph(_=None):
        data = update_data(ax, n_slider.val)
        N_points = data.shape[0]
        ax.set(xlim3d=(0, N_points - 1))
        ax.set(ylim3d=(0, N_points - 1))
        ax.set(zlim3d=(0, N_points - 1))

        wmax = np.max(np.abs(data))
        threshold = wmax / threshold_fact
        data = data[:, N_points:, N_points:]
        mask = np.abs(data) >= threshold
        relevant = np.argwhere(mask)
        data_c = (data[mask] - threshold) / (wmax - threshold)

        graph._offsets3d = np.array([relevant[:, 0], relevant[:, 1], relevant[:, 2]])
        graph.set_array(data_c)

    ## Sliders & Buttons
    fig.subplots_adjust(bottom=0.2)
    vmin, vmax, vstep, slidertitle, vformat = slider_params
    n_slider = Slider(ax=fig.add_axes((0.25, 0.1, 0.65, 0.03)), label=slidertitle, valfmt=vformat, valmin=vmin, valmax=vmax, valinit=vmin, valstep=vstep)
    n_slider.on_changed(update_graph)

    update_graph()

    return n_slider, ax


# Other
def rand_cmap(
    nlabels: int, type_: typing.Literal["bright", "soft"] = "bright", first_color_black: bool = True, last_color_black: bool = False
) -> "LinearSegmentedColormap":
    """Creates a random colormap to be used together with matplotlib. Useful for segmentation tasks

    Args:
        nlabels: Number of labels (size of colormap)
        type_: 'bright' for strong colors, 'soft' for pastel colors
        first_color_black: Option to use first color as black
        last_color_black: Option to use last color as black

    Returns:
        colormap for matplotlib
    """
    import colorsys

    from pyloggrid.Libs.singlethread_numpy import np

    np.zeros(0)  # Do not remove - used to force singlethread_numpy to be imported before the other libraries when refactoring

    from matplotlib.colors import LinearSegmentedColormap

    if type_ not in ("bright", "soft"):
        raise ValueError('Please choose "bright" or "soft" for type')

    # Generate color map for bright colors, based on hsv
    if type_ == "bright":
        randHSVcolors = [(np.random.uniform(low=0.0, high=1), np.random.uniform(low=0.2, high=1), np.random.uniform(low=0.9, high=1)) for _ in range(nlabels)]

        randRGBcolors = [colorsys.hsv_to_rgb(HSVcolor[0], HSVcolor[1], HSVcolor[2]) for HSVcolor in randHSVcolors]
        if first_color_black:
            randRGBcolors[0] = [0, 0, 0]

        if last_color_black:
            randRGBcolors[-1] = [0, 0, 0]

        random_colormap = LinearSegmentedColormap.from_list("new_map", randRGBcolors, N=nlabels)

    elif type_ == "soft":
        low = 0.6
        high = 0.95
        randRGBcolors: list = [
            (np.random.uniform(low=low, high=high), np.random.uniform(low=low, high=high), np.random.uniform(low=low, high=high)) for _ in range(nlabels)
        ]

        if first_color_black:
            randRGBcolors[0] = [0, 0, 0]

        if last_color_black:
            randRGBcolors[-1] = [0, 0, 0]
        random_colormap = LinearSegmentedColormap.from_list("new_map", randRGBcolors, N=nlabels)
    else:
        raise ValueError

    return random_colormap


initFormat()



"""Cython settings for linux"""

import os
from distutils.core import setup
from distutils.extension import Extension

import numpy as np
from Cython.Build import cythonize

os.environ["CC"] = "clang"
os.environ["CXX"] = "clang -shared"

ext_modules = [
    Extension(
        "compute_convolution_kernel",
        ["compute_convolution_kernel.pyx"],
        extra_compile_args=["-fopenmp", "-march=native", "-c", "-g", "-Wextra", "-march=native", "-ffast-math", "-funroll-loops", "-fno-stack-protector"],
        extra_link_args=["-fopenmp"],
    )
]

setup(
    ext_modules=cythonize(ext_modules, annotate=True),
    include_dirs=[np.get_include()],
    name="log-grid",
    author="A. Barral",
)



# cython: language_level=3, boundscheck=False, wraparound=False, initializedcheck=False, cdivision=True, infer_types=True
import numpy as np
cimport numpy as np
from libc.stdint cimport uint32_t

# python setup_linux.py build_ext --inplace

cdef inline short offset_coord(const short pos, const short offset, const Py_ssize_t N_pts, const bint is_axis_0, const bint has_k0) nogil:
    """offset a coord, returns the offset coord or -1 if out of bounds"""
    cdef short coord
    if is_axis_0:
        coord = pos + offset
        if has_k0:
            oob = not((0 < coord < N_pts) or (pos==0 and coord==0))
        else:
            oob = not (0 <= coord < N_pts)
    else:
        n = N_pts//2
        if pos < n:
            coord = pos - offset
            oob = not (0 <= coord < n)
        else:
            coord = pos + offset
            if has_k0:
                oob = not ((n < coord < N_pts) or (pos == n and coord == n))
            else:
                oob = not (n <= coord < N_pts)
    if oob:
        coord = -1

    return coord

cdef inline short flip_coord(short coord, Py_ssize_t N_pts, bint has_k0) nogil:
    """flip a coordinate"""
    cdef short n, n_k0
    n = <short>(N_pts // 2)
    n_k0 = (0 if has_k0 else 1)
    if coord < n:
        coord = n + n - coord - n_k0
    else:
        coord = n - (coord - n + n_k0)
    return coord

cdef inline Py_ssize_t midx_2d(const short a, const short b, const short ny) nogil:
    """3D to 1D coord"""
    return (a * ny) + b

cdef inline Py_ssize_t midx_3d(const short a, const short b, const short c, const short ny, const short nz) nogil:
    """3D to 1D coord"""
    return (a * nz + b) * ny + c

cdef np.ndarray compute_interaction_kernel_1D_c(short[:, ::1] offsets_v, short[:, ::1] signs_v, short[:, ::1] offsets_k0_v, short[:, ::1] signs_k0_v, const bint has_k0, const short sx):
    cdef short sign_x0, sign_x1, coord_x0, coord_x1
    cdef uint32_t kernel_i = -1
    cdef short[:, ::1] offsets_i_v, signs_i_v
    kernel_size = sx * offsets_v.shape[0]
    if has_k0:
        kernel_size *= 2
    cdef np.ndarray kernel = np.zeros((kernel_size, 4), dtype=np.uint32)
    cdef uint32_t[:, ::1] kernel_v = kernel
    for i_nx in range(sx):
        if has_k0 and i_nx == 0:
            offsets_i_v = offsets_k0_v
            signs_i_v = signs_k0_v
        else:
            offsets_i_v = offsets_v
            signs_i_v = signs_v
        for i_kx in range(offsets_i_v.shape[0]):
            sign_x0, sign_x1 = signs_i_v[i_kx][0], signs_i_v[i_kx][1]
            coord_x0 = offset_coord(i_nx, offsets_i_v[i_kx][0], sx, True, has_k0)
            coord_x1 = offset_coord(i_nx, offsets_i_v[i_kx][1], sx, True, has_k0)
            if coord_x0 < 0 or coord_x1 < 0:  # check that both iteractions are in the grid
                continue
            kernel_i += 1

            kernel_v[kernel_i, 0] = i_nx
            if sign_x0 == -1:
                kernel_v[kernel_i, 1] = coord_x0
                kernel_v[kernel_i, 2] = coord_x1
                kernel_v[kernel_i, 3] = 1  # 1: f.c * g
            elif sign_x1 == -1:
                kernel_v[kernel_i, 1] = coord_x0
                kernel_v[kernel_i, 2] = coord_x1
                kernel_v[kernel_i, 3] = 2  # 2: f * g.c
            else:
                kernel_v[kernel_i, 1] = coord_x0 * sign_x0
                kernel_v[kernel_i, 2] = coord_x1 * sign_x1
                kernel_v[kernel_i, 3] = 0  # 0: f * g
    return kernel[:kernel_i+1]

cpdef np.ndarray compute_interaction_kernel_1D(short[:, ::1] offsets_v, short[:, ::1] signs_v, short[:, ::1] offsets_k0_v, short[:, ::1] signs_k0_v, const bint has_k0, const short sx):
    return compute_interaction_kernel_1D_c(offsets_v, signs_v, offsets_k0_v, signs_k0_v, has_k0, sx).flatten()

cdef np.ndarray compute_interaction_kernel_2D_c(short[:, ::1] offsets_v, short[:, ::1] signs_v, short[:, ::1] offsets_k0_v, short[:, ::1] signs_k0_v, short[:, ::1] offsets_k1_v, short[:, ::1] signs_k1_v, const bint has_k0, const short sx, const short sy):
    cdef short sign_x0, sign_x1, sign_y0, sign_y1, coord_x0, coord_x1, coord_y0, coord_y1
    cdef uint32_t kernel_i = -1
    cdef short[:, ::1] offsets_x_v, offsets_y_v, signs_x_v, signs_y_v
    kernel_size = sx * sy * offsets_v.shape[0] **2
    if has_k0:
        kernel_size *= 2
    cdef np.ndarray kernel = np.zeros((kernel_size, 4), dtype=np.uint32)
    cdef uint32_t[:, ::1] kernel_v = kernel
    for i_nx in range(sx):
        if has_k0 and i_nx == 0:
            offsets_x_v = offsets_k0_v
            signs_x_v = signs_k0_v
        else:
            offsets_x_v = offsets_v
            signs_x_v = signs_v
        for i_kx in range(offsets_x_v.shape[0]):
            sign_x0, sign_x1 = signs_x_v[i_kx][0], signs_x_v[i_kx][1]
            coord_x0 = offset_coord(i_nx, offsets_x_v[i_kx][0], sx, True, has_k0)
            coord_x1 = offset_coord(i_nx, offsets_x_v[i_kx][1], sx, True, has_k0)
            if coord_x0 < 0 or coord_x1 < 0:  # check that both iteractions are in the grid
                continue

            for i_ny in range(sy):  # Y
                if has_k0 and i_ny == sy//2:
                    offsets_y_v = offsets_k1_v
                    signs_y_v = signs_k1_v
                else:
                    offsets_y_v = offsets_v
                    signs_y_v = signs_v
                for i_ky in range(offsets_y_v.shape[0]):
                    sign_y0, sign_y1 = signs_y_v[i_ky][0], signs_y_v[i_ky][1]
                    coord_y0 = offset_coord(i_ny, offsets_y_v[i_ky][0], sy , False, has_k0)
                    coord_y1 = offset_coord(i_ny, offsets_y_v[i_ky][1], sy, False, has_k0)
                    if coord_y0 < 0 or coord_y1 < 0:  # check that both iteractions are in the grid
                        continue
                    kernel_i += 1

                    # remap
                    if sign_y0 == -1:
                        coord_y0 = flip_coord(coord_y0, sy, has_k0)
                    elif sign_y1 == -1:
                        coord_y1 = flip_coord(coord_y1, sy, has_k0)
                    elif sign_y0 == 0:
                        coord_y0 = int(sy // 2)
                    elif sign_y1 == 0:
                        coord_y1 = int(sy // 2)

                    kernel_v[kernel_i, 0] = midx_2d(i_nx, i_ny, sy)
                    if sign_x0 == -1:
                        kernel_v[kernel_i, 1] = midx_2d(coord_x0, flip_coord(coord_y0, sy, has_k0), sy)
                        kernel_v[kernel_i, 2] = midx_2d(coord_x1, coord_y1, sy)
                        kernel_v[kernel_i, 3] = 1  # 1: f.c * g
                    elif sign_x1 == -1:
                        kernel_v[kernel_i, 1] = midx_2d(coord_x0, coord_y0, sy)
                        kernel_v[kernel_i, 2] = midx_2d(coord_x1, flip_coord(coord_y1, sy, has_k0), sy)
                        kernel_v[kernel_i, 3] = 2  # 2: f * g.c
                    else:
                        kernel_v[kernel_i, 1] = midx_2d(coord_x0 * sign_x0, coord_y0, sy)
                        kernel_v[kernel_i, 2] = midx_2d(coord_x1 * sign_x1, coord_y1, sy)
                        kernel_v[kernel_i, 3] = 0  # 0: f * g
    return kernel[:kernel_i+1]

cpdef np.ndarray compute_interaction_kernel_2D(short[:, ::1] offsets_v, short[:, ::1] signs_v, short[:, ::1] offsets_k0_v, short[:, ::1] signs_k0_v, short[:, ::1] offsets_k1_v, short[:, ::1] signs_k1_v, const bint has_k0, const short sx, const short sy):
    return compute_interaction_kernel_2D_c(offsets_v, signs_v, offsets_k0_v, signs_k0_v, offsets_k1_v, signs_k1_v, has_k0, sx, sy).flatten()

cdef np.ndarray compute_interaction_kernel_3D_c(short[:, ::1] offsets_v, short[:, ::1] signs_v, short[:, ::1] offsets_k0_v, short[:, ::1] signs_k0_v, short[:, ::1] offsets_k1_v, short[:, ::1] signs_k1_v, const bint has_k0, const short sx, const short sy, const short sz):
    cdef short sign_x0, sign_x1, sign_y0, sign_y1, sign_z0, sign_z1, coord_x0, coord_x1, coord_y0, coord_y1, coord_z0, coord_z1
    cdef short[:, ::1] offsets_x_v, offsets_y_v, offsets_z_v, signs_x_v, signs_y_v, signs_z_v
    cdef uint32_t kernel_i = -1
    kernel_size = sx * sy * sz * offsets_v.shape[0] ** 3
    if has_k0:
        kernel_size *= 2
    cdef np.ndarray kernel = np.zeros((kernel_size, 4), dtype=np.uint32)
    cdef uint32_t[:, ::1] kernel_v = kernel
    for i_nx in range(sx):
        if has_k0 and i_nx == 0:
            offsets_x_v = offsets_k0_v
            signs_x_v = signs_k0_v
        else:
            offsets_x_v = offsets_v
            signs_x_v = signs_v
        for i_kx in range(offsets_x_v.shape[0]):
            sign_x0, sign_x1 = signs_x_v[i_kx][0], signs_x_v[i_kx][1]
            coord_x0 = offset_coord(i_nx, offsets_x_v[i_kx][0], sx, True, has_k0)
            coord_x1 = offset_coord(i_nx, offsets_x_v[i_kx][1], sx, True, has_k0)
            if coord_x0 < 0 or coord_x1 < 0:  # check that both iteractions are in the grid
                continue

            for i_ny in range(sy):  # Y
                if has_k0 and i_ny == sy//2:
                    offsets_y_v = offsets_k1_v
                    signs_y_v = signs_k1_v
                else:
                    offsets_y_v = offsets_v
                    signs_y_v = signs_v
                for i_ky in range(offsets_y_v.shape[0]):
                    sign_y0, sign_y1 = signs_y_v[i_ky][0], signs_y_v[i_ky][1]
                    coord_y0 = offset_coord(i_ny, offsets_y_v[i_ky][0], sy , False, has_k0)
                    coord_y1 = offset_coord(i_ny, offsets_y_v[i_ky][1], sy, False, has_k0)
                    if coord_y0 < 0 or coord_y1 < 0:  # check that both iteractions are in the grid
                        continue

                    # remap
                    if sign_y0 == -1:
                        coord_y0 = flip_coord(coord_y0, sy, has_k0)
                    elif sign_y1 == -1:
                        coord_y1 = flip_coord(coord_y1, sy, has_k0)
                    elif sign_y0 == 0:
                        coord_y0 = int(sy // 2)
                    elif sign_y1 == 0:
                        coord_y1 = int(sy // 2)

                    for i_nz in range(sz):  # Z
                        if has_k0 and i_nz == sz // 2:
                            offsets_z_v = offsets_k1_v
                            signs_z_v = signs_k1_v
                        else:
                            offsets_z_v = offsets_v
                            signs_z_v = signs_v
                        for i_kz in range(offsets_z_v.shape[0]):
                            sign_z0, sign_z1 = signs_z_v[i_kz][0], signs_z_v[i_kz][1]
                            coord_z0 = offset_coord(i_nz, offsets_z_v[i_kz][0], sz, False, has_k0)
                            coord_z1 = offset_coord(i_nz, offsets_z_v[i_kz][1], sz, False, has_k0)
                            if coord_z0 < 0 or coord_z1 < 0:  # check that both iteractions are in the grid
                                continue
                            kernel_i += 1

                            # remap
                            if sign_z0 == -1:
                                coord_z0 = flip_coord(coord_z0, sy, has_k0)
                            elif sign_z1 == -1:
                                coord_z1 = flip_coord(coord_z1, sy, has_k0)
                            elif sign_z0 == 0:
                                coord_z0 = int(sz // 2)
                            elif sign_z1 == 0:
                                coord_z1 = int(sz // 2)

                            kernel_v[kernel_i,0] = midx_3d(i_nx, i_ny, i_nz, sy, sz)
                            if sign_x0 == -1:
                                kernel_v[kernel_i, 1] = midx_3d(coord_x0, flip_coord(coord_y0, sy, has_k0), flip_coord(coord_z0, sz, has_k0), sy, sz)
                                kernel_v[kernel_i, 2] = midx_3d(coord_x1, coord_y1, coord_z1, sy, sz)
                                kernel_v[kernel_i, 3] = 1  # 1: f.c * g
                            elif sign_x1 == -1:
                                kernel_v[kernel_i, 1] = midx_3d(coord_x0, coord_y0, coord_z0, sy, sz)
                                kernel_v[kernel_i, 2] = midx_3d(coord_x1, flip_coord(coord_y1, sy, has_k0), flip_coord(coord_z1, sz, has_k0), sy, sz)
                                kernel_v[kernel_i, 3] = 2  # 2: f * g.c
                            else:
                                kernel_v[kernel_i, 1] = midx_3d(coord_x0 * sign_x0, coord_y0, coord_z0, sy, sz)
                                kernel_v[kernel_i, 2] = midx_3d(coord_x1 * sign_x1, coord_y1, coord_z1, sy, sz)
                                kernel_v[kernel_i, 3] = 0  # 0: f * g

    return kernel[:kernel_i+1]

cpdef np.ndarray compute_interaction_kernel_3D(short[:, ::1] offsets_v, short[:, ::1] signs_v, short[:, ::1] offsets_k0_v, short[:, ::1] signs_k0_v, short[:, ::1] offsets_k1_v, short[:, ::1] signs_k1_v, const bint has_k0, const short sx, const short sy, const short sz):
    return compute_interaction_kernel_3D_c(offsets_v, signs_v, offsets_k0_v, signs_k0_v, offsets_k1_v, signs_k1_v, has_k0, sx, sy, sz).flatten()



"""Cython settings for windows"""

from distutils.core import setup

import numpy
from Cython.Build import cythonize
from setuptools import Extension

ext_modules = [
    Extension(
        "compute_convolution_kernel",
        ["compute_convolution_kernel.pyx"],
        extra_compile_args=["/openmp"],
    )
]
setup(
    ext_modules=cythonize(ext_modules, annotate=True, compiler_directives={"language_level": "3"}),
    include_dirs=[numpy.get_include()],
    name="log-grid",
    author="A. Barral",
)



"""Create a custom logger. Used to replace all print calls. Outputs to log file."""

import logging
import sys


def setup_custom_logger(name: str, level=logging.DEBUG, logfile: str = "log.log") -> logging.Logger:
    """Create custom logger. Auto-formats logs lines, writes to file.

    Args:
        logfile: path to log file to write to
        name: logger name
        level: log level (debug/info/..)

    Returns:
        logger
    """
    formatter = logging.Formatter(fmt="%(asctime)s %(levelname)-8s %(name)-20s %(message)s", datefmt="%Y-%m-%d %H:%M:%S")
    handler = logging.FileHandler(logfile, "a", encoding="utf-8")
    handler.setFormatter(formatter)
    screen_handler = logging.StreamHandler(stream=sys.stdout)
    screen_handler.setFormatter(formatter)
    logger = logging.getLogger(name)
    logger.setLevel(level)
    for hand in logger.handlers[:]:  # remove logger if already exists (jupyter notebook, ...)
        logger.removeHandler(hand)
    logger.addHandler(handler)
    logger.addHandler(screen_handler)
    logger.propagate = False
    return logger
